// photo.cxx -- main loop
//
// Written by Curtis Olson, started May 2001.
//
// Copyright (C) 2001  Curtis L. Olson  - http://www.flightgear.org/~curt
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: photo.cxx,v 1.15 2004-11-19 22:25:51 curt Exp $

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <simgear/compiler.h>

#include <stdio.h>
#include <stdlib.h>

#include <string>
#include <iostream>

#include <simgear/bucket/newbucket.hxx>
#include <simgear/debug/logstream.hxx>
#include <simgear/io/sg_binobj.hxx>
#include <simgear/math/SGGeodesy.hxx>
#include <simgear/math/SGGeometry.hxx>
#include <simgear/math/SGMath.hxx>

#include <Array/array.hxx>
#include <Geometry/trinodes.hxx>
#include <Output/output.hxx>
#include <Polygon/chop.hxx>
#include <Polygon/index.hxx>
#include <Polygon/polygon.hxx>

using std::string;
using std::cout;
using std::endl;


const int MAX_XDIV = 16;
const int MAX_YDIV = 16;


// fix node elevations
point_list calc_elevations( const string& root, const point_list& geod_nodes ) {
    bool done = false;
    point_list result = geod_nodes;
    int i, j;
    TGArray array;

    // set all elevations to -9999
    for ( i = 0; i < (int)result.size(); ++i ) {
	result[i].setz( -9999.0 );
    }

    while ( !done ) {
	// find first node with -9999 elevation
	i = 0;
	while ( (result[i].z() > -9000) && (i < (int)result.size()) ) {
	    ++i;
	}

	if ( i < (int)result.size() ) {
	    SGBucket b( result[i].x(), result[i].y() );
	    string base = b.gen_base_path();

	    // try 3 arcsec dems first
	    string dem_path = root + "/DEM-3/" + base 
		+ "/" + b.gen_index_str();
	    cout << "dem_path = " << dem_path << endl;
	
	    if ( ! array.open(dem_path) ) {
		cout << "ERROR: cannot open 3 arcsec file " << dem_path << endl;
		cout << "trying 30 arcsec file" << endl;
		
		// try 30 arcsec dem
		dem_path = root + "/DEM-30/" + base 
		    + "/" + b.gen_index_str();
		cout << "dem_path = " << dem_path << endl;
		if ( ! array.open(dem_path) ) {
		    cout << "ERROR: cannot open 3 arcsec file " 
			 << dem_path << endl;
		}
	    }
	    array.parse( b );

	    // update all the non-updated elevations that are inside
	    // this dem file
	    double elev;
	    done = true;
	    for ( j = 0; j < (int)result.size(); ++j ) {
		if ( result[j].z() < -9000 ) {
		    done = false;
		    cout << "interpolating for " << result[j] << endl;
		    elev = array.altitude_from_grid( result[j].x() * 3600.0,
                                                     result[j].y() * 3600.0 );
		    if ( elev > -9000 ) {
			result[j].setz( elev );
		    }
		}
	    }
	    array.close();
	} else {
	    done = true;
	}
    }


    return result;
}


int main( int argc, char **argv ) {
    int i, j;

    sglog().setLogLevels( SG_ALL, SG_DEBUG );

    // check args
    if ( argc != 13 ) {
        SG_LOG( SG_GENERAL, SG_ALERT,
                "Usage " << argv[0]
                << " root image(base) xdiv ydiv x0 y0 x1 y1 x2 y2 x3 y3" );
        exit(-1);
    }

    string root = argv[1];
    string image = argv[2];

    int xdiv = atoi( argv[3] );
    int ydiv = atoi( argv[4] );
    if ( xdiv > MAX_XDIV || xdiv < 1 || ydiv > MAX_YDIV || ydiv < 1 ) {
        SG_LOG( SG_GENERAL, SG_ALERT,
                "{x,y}div must be in the range of 1 - " << MAX_XDIV );
        exit(-1);
    }
        
    double x0 = atof( argv[5] );
    double y0 = atof( argv[6] );

    double x1 = atof( argv[7] );
    double y1 = atof( argv[8] );

    double x2 = atof( argv[9] );
    double y2 = atof( argv[10] );

    double x3 = atof( argv[11] );
    double y3 = atof( argv[12] );

    // generate vertices

    TGTriNodes nodes; nodes.clear();

    double dx0 = (x3 - x0) / ydiv;
    double dy0 = (y3 - y0) / ydiv;
    double dx1 = (x2 - x1) / ydiv;
    double dy1 = (y2 - y1) / ydiv;

    double xarray[32][32];
    double yarray[32][32];
    for ( j = 0; j <= ydiv; ++j ) {
        double tx0 = x0 + j * dx0;
        double ty0 = y0 + j * dy0;
        double tx1 = x1 + j * dx1;
        double ty1 = y1 + j * dy1;
        double dx = (tx1 - tx0) / xdiv;
        double dy = (ty1 - ty0) / xdiv;
        for ( i = 0; i <= xdiv; ++i ) {
            xarray[i][j] = tx0 + i * dx;
            yarray[i][j] = ty0 + i * dy;
            cout << "(" << xarray[i][j] << "," << yarray[i][j] << ")" << endl;
            nodes.simple_add( Point3D(xarray[i][j], yarray[i][j], 0) );
        }
        cout << endl;
    }

    // create the object structures

    // geodetic nodes
    point_list geod_nodes = calc_elevations( root, nodes.get_node_list() );

    // texture coordinates
    TGTriNodes texcoords; texcoords.clear();
    texcoords.simple_add( Point3D( 0.0, 0.0, 0.0 ) );
    texcoords.simple_add( Point3D( 1.0, 0.0, 0.0 ) );
    texcoords.simple_add( Point3D( 1.0, 1.0, 0.0 ) );
    texcoords.simple_add( Point3D( 0.0, 1.0, 0.0 ) );

    // triangles
    group_list strips_v; strips_v.clear();
    group_list strips_tc; strips_tc.clear();
    string_list strip_materials; strip_materials.clear();
    int_list strip_v;
    int_list strip_tc;

    int count = 0;
    for ( j = 0; j < ydiv; ++j ) {
        for ( i = 0; i < xdiv; ++i ) {
            strip_v.clear();
            strip_v.push_back( count );
            strip_v.push_back( count + 1 );
            strip_v.push_back( count + xdiv + 1 );
            strip_v.push_back( count + xdiv + 2 );

            strip_tc.clear();
            strip_tc.push_back( 0 );
            strip_tc.push_back( 1 );
            strip_tc.push_back( 3 );
            strip_tc.push_back( 2 );

            char bufx[5], bufy[5];
            snprintf( bufx, 5, "%X", i );
            snprintf( bufy, 5, "%X", j );
            string material = image;
            material += bufx;
            material += bufy;
            material += ".rgb";

            strips_v.push_back( strip_v );
            strips_tc.push_back( strip_tc );
            strip_materials.push_back( material );

            ++count;
        }
        ++count;
    }

    // wgs84 cartesian nodes
    std::vector< SGVec3d > wgs84_nodes; wgs84_nodes.clear();
    for ( i = 0; i < (int)geod_nodes.size(); ++i ) {
        SGGeod geod;
        geod = SGGeod::fromDegM( geod_nodes[i].x(), geod_nodes[i].y(), geod_nodes[i].z() );
        SGVec3d cart = SGVec3d::fromGeod(geod);
        cout << cart << endl;
	wgs84_nodes.push_back( cart );
    }

    // bounding sphere
    SGBucket b( (x0 + x2) / 2, (y0 + y2) / 2 );
    SGSphered d;
    for ( i = 0; i < wgs84_nodes.size(); ++i ) {
        d.expandBy(wgs84_nodes[ i ]);
    }
    
    SGVec3d gbs_center = d.getCenter();
    double gbs_radius = d.getRadius();
    cout << "gbs center = " << gbs_center << endl;

    // normals
    std::vector< SGVec3f > normals;
    for ( i = 0; i < (int)wgs84_nodes.size(); ++i ) {
        normals.push_back(toVec3f(normalize( wgs84_nodes[i] )));
    }
    
    const point_list &tc_nodes = texcoords.get_node_list();
    std::vector< SGVec2f > texcoords_vec;
    for ( i = 0; i < (int)tc_nodes.size(); ++i )
    {
        texcoords_vec.push_back( SGVec2f( tc_nodes[i].x(), tc_nodes[i].y() ) );
    }

    // build the object
    SGBinObject obj;

    // null structures
    group_list tris_v; tris_v.clear();
    group_list tris_tc; tris_tc.clear();
    string_list tri_materials; tri_materials.clear();

    group_list fans_v; fans_v.clear();
    group_list fans_tc; fans_tc.clear();
    string_list fan_materials; fan_materials.clear();

    obj.set_gbs_center( gbs_center );
    cout << "gbs center = " << gbs_center << endl;
    obj.set_gbs_radius( gbs_radius );
    obj.set_wgs84_nodes( wgs84_nodes );
    obj.set_normals( normals );
    obj.set_texcoords( texcoords_vec );
    obj.set_tris_v( tris_v );
    obj.set_tris_tc( tris_tc ); 
    obj.set_tri_materials( tri_materials );
    obj.set_strips_v( strips_v );
    obj.set_strips_tc( strips_tc ); 
    obj.set_strip_materials( strip_materials );
    obj.set_fans_v( fans_v );
    obj.set_fans_tc( fans_tc );
    obj.set_fan_materials( fan_materials );

    // write the object
    string objpath = root + "/PhotoObj";
    string name = image + ".btg";

    bool result = obj.write_bin( objpath, name, b );
    if ( !result ) {
	cout << "error writing file. :-(" << endl;
	exit(-1);
    }

    // write the index entry
    write_index( objpath, b, name );

    // create and write the 'hole' polygon
    TGPolygon hole; hole.erase();
    Point3D p;

    for ( i = 0; i <= xdiv; ++i ) {
        p = Point3D( xarray[i][0], yarray[i][0], 0 );
        hole.add_node( 0, p );
    }
    for ( j = 0; j <= ydiv; ++j ) {
        p = Point3D( xarray[xdiv][j], yarray[xdiv][j], 0 );
        hole.add_node( 0, p );
    }
    for ( i = xdiv; i >= 0; --i ) {
        p = Point3D( xarray[i][ydiv], yarray[i][ydiv], 0 );
        hole.add_node( 0, p );
    }
    for ( j = ydiv; j >= 0; --j ) {
        p = Point3D( xarray[0][j], yarray[0][j], 0 );
        hole.add_node( 0, p );
    }

    // initialize persistant polygon counter
    string counter_file = root + "/poly_counter";
    poly_index_init( counter_file );

    string holepath = root + "/PhotoArea";
    tgChopNormalPolygon( holepath, "Hole", hole, false );

    return 0;
}
